/***********************************************************************
 * File: days.h
 * 	(Extracted from
 *	   "$Id: days.h,v 1.1 2007/10/29 06:55:04 Administrator Exp Administrator $")
 *
 * Abstracts:
 *  This C header file defines the is_valid_date() and date_in_English()
 * API-s.
 *
 * Copyright (C) 2005-2007 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#ifndef _DAYS_H_ /* { */
#define _DAYS_H_
#ifdef __cplusplus
extern "C" {
#endif

/* include the system definitions */
#include <unistd.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <limits.h>
#include <errno.h>

/* for executing "gcc -std=c99..." */
extern int snprintf(char *STR, size_t SIZE, const char *FORMAT,...);
extern long long llabs(long long j);

/* The default year of these API is 2000A.D. */
#define DAYS_BASE_YEAR		2000UL

#define IS_LEAP_YEAR(year) (					  \
			((year) % 400 == 0)			  \
			|| ((year) % 100 != 0 && (year) % 4 == 0) \
		)

/*
 * This API determines the numeric string is valid date (yyy...yyymmdd)
 *
 * arguments:
 *	date is pointing to the date string (yyy...yyymmdd).
 *	sz indicates the size of date pointing area.
 *
 * returns:
 *	DAYS_ILLEGAL_ARGS (= -1) : illegal arguments
 *	DAYS_FALSE (= 0)         : the string doesn't means valid date.
 *	DAYS_TRUE  (= 1)         : the string means valid date.
 */
extern int is_valid_date(char *date, size_t sz);
#	define	DAYS_ILLEGAL_ARGS	-1
#	define	DAYS_FALSE		0
#	define	DAYS_TRUE		1

/*
 * This API gets the year, month and day from a date string.
 *
 * arguments:
 *	 date points to the numeric string.
 *	 sz indicates the length of date strings.
 *	 year points to the integer as year.
 *	 month points to the integer as month.
 *	 day points to the integer as day.
 *
 * returns:
 *	DAYS_ILLEGAL_ARGS : it means illegal arguments.
 *	0 : normal end
 */
extern int get_ymd(char		*date
		, size_t	sz
		, unsigned long	*year
		, int		*month
		, int		*day
		);

/*
 * This API translate the numeric string into date in English.
 *
 * arguments:
 *	date is pointing to the numeric string (yyy...yyymmdd).
 *	sz indicates the size of str pointing area.
 *	fmt indicates the format of European or American.
 *	str is pointing the output string.
 *	str_sz indicates the size of the str output area.
 *
 * returns:
 *	NULL : illegal arguments
 *	non-NULL : the translated string address (same as str)
 */
/* type definition for date_in_English() API */
typedef enum days_fmt_t {
	  DAYS_FMT_EUROPEAN
	, DAYS_FMT_AMERICAN
	}	DAYS_FMT;

extern char *date_in_English(char *date
		, size_t sz
		, DAYS_FMT fmt
		, char *str
		, size_t str_sz
		);

/*
 * This API returns the day of year.
 *
 * arguments:
 *	date is pointing to the date string.
 *	sz indicates the length of the date string.
 *
 * returns:
 *	zero : illegal arguments
 *	positive integers: the Julian date for date string
 */
int get_day_of_year(char *date, size_t sz);

/*
 * This API counts the days since 1, January AD.1 (it's the 1st day).
 *
 * arguments:
 *	date is pointing to the date.
 *	sz indicates the length of the date string.
 *
 * returns:
 *	LLONG_MIN :		illegal arguments
 *	0LL to 1568704592244LL:	the count of days.
 */
extern long long days_count(char *date, size_t sz);
#define MAX_DAYS_COUNT 1568704592244LL /* as "42949672951231" date */

/*
 * Macro `day_of_week()' returns indicator as the day of week.
 * API `day_of_week_str()' returns pointer to name of the day of week.
 *
 * arguments:
 *	days_counts is returns of `days_count()' API.
 *
 * returns (`day_of_week()' macro):
 *	Illegal_days_counts		: illegal arguments
 *	Sunday (== 0) to Saturday (== 6): indicates day of week
 * returns (`day_of_week_str()' API):
 *	NULL	: illegal arguments
 *	non-NULL: address of the name of the day of week.
 */
typedef enum day_of_week_e {
	  Sunday	= 0
	, Monday
	, Tuesday
	, Wednesday
	, Thursday
	, Friday
	, Saturday
	, Illegal_day_of_week /* it must be greater than zero */
} Day_of_week_e;

#define day_of_week(Days_count) (Day_of_week_e)(		\
		((Days_count) >= 0LL)?				\
			((Days_count) % Illegal_day_of_week)	\
		:	Illegal_day_of_week			\
	)

extern char *day_of_week_str(long long days_counts);

/*
 * This API returns pointer to the name of the day of week
 * from a date string.
 *
 * arguments:
 *	date points to the string of a date.
 *	sz indicates the length of `date'.
 *
 * returns:
 *	NULL	: illegal arguments
 *	non-NULL: address of the name of the day of week.
 */
extern char *get_day_of_week(char *date, size_t sz);

/*
 * This API subtracts the day-counts of date1 and date2.
 *
 * arguments:
 *	date1 is pointing to the subtracters date string.
 *	sz1 indicates the length of date1 staring.
 *	date2 is pointing to the subtracter date string.
 *	sz2 indicates the length of date2 staring.
 *
 * returns:
 *	LLONG_MIN : illegal arguments
 *	-1568704592244LL to 1568704592244LL: the diff of day-counts.
 */
extern long long diff_dates(char *date1
			, size_t sz1
			, char *date2
			, size_t sz2
			);

/*
 * This API calculate year, month and day from result of `days_count()'.
 *
 * arguments:
 *	day_count indicates the result of `days_count()' API.
 *	year points to the unsigned long as years of `day_count'.
 *	month points to the integer as month of `day_count'.
 *	day points to the integer as day of `day_count'.
 *
 * returns:
 *	DAYS_ILLEGAL_ARGS (= -1) : illegal arguments
 *	0			 : normal end
 */
extern int days_count_to_ymd(long long day_count
				, unsigned long *year
				, int *month
				, int *day
				);

/*
 * This API translate day-count to the 'YYY...YYYMMDD' formatted string.
 *
 * arguments:
 *	  day_count indicates the count of past days that returns of
 *	`days_count()' API.
 *	  date points to the buffer that is outputted 'YYY...YYYMMDD'
 *	formatted string.
 *	  sz indicates the size of buffer that is pointed from `date'.
 *	sz must be greater than nine bytes.
 *
 * returns:
 *	NULL: invalid arguments
 *	non-NULL: normal end (same as `date')
 */
extern char *days_count_to_date_str(long long day_count
				, char *date
				, size_t sz
				);

/*
 * This API calculate durations for years, months, weeks.
 *
 * arguments:
 *	 date points to a numeral string.
 *	 unit indicates the duration unit (Duration_unit_e).
 *	 delta indicates the delta for date argument.
 *	 duration points to the result buffer of calculating duration.
 *
 * returns:
 *	NULL: invalid arguments
 *	non-NULL: same as duration argument
 */
typedef struct duration_t {
	long long	begin_of, end_of, days;
} Duration_t;

/* This macro predicates the Duration_t is illegal or not. */
#define Is_illegal_Duration_t(Dur) (((				\
		(Duration_t *)(Dur) == NULL)			\
		|| ((Duration_t *)(Dur))->days < 0)		\
		|| ((Duration_t *)(Dur))->begin_of >		\
			((Duration_t *)(Dur))->end_of		\
	)

typedef enum duration_unit_e {
	  Duration_unit_Years
	, Duration_unit_Months
	, Duration_unit_Weeks
} Duration_unit_e;

extern Duration_t *calc_duration(char			*date
				, Duration_unit_e	unit
				, long long		delta
				, Duration_t		*duration
				);

/*
 * This API calculate durations for years, months, weeks
 * from a result of `days_count()'.
 *
 * arguments:
 *	 day_count indicates the result of `days_count()' API.
 *	 unit indicates the duration unit (Duration_unit_e).
 *	 delta indicates the delta for date argument.
 *	 duration points to the result buffer of calculating duration.
 *
 * returns:
 *	NULL: invalid arguments
 *	non-NULL: same as duration argument
 */
extern Duration_t *get_duration(long long		day_count
				, Duration_unit_e	unit
				, long long		delta
				, Duration_t		*duration
				);

/*
 * This API sets the begin and end of a duration the Duration_t.
 *
 * arguments:
 *	 dur points to the Duration_t to set.
 *	 begin_of indicates the begin of a duration.
 *	 end_of indicates the end of a duration.
 *
 * returns:
 *	NULL		: illegal arguments
 *	non-NULL	: normal end (same as dur)
 */
extern Duration_t *set_duration(Duration_t *dur
				, long long begin_of
				, long long end_of
				);

/*
 * This API compares two durations, and calculates closure and
 * intersection from those.
 *
 * arguments:
 *	 dur1 indicates the left-hand data to compare.
 *	 dur2 indicates the right-hand data to compare.
 *	 cmp_result is pointer to the result of comparing.  if it is -2,
 *	then dur1 and dur2 has no intersections and then dur1 is on the
 *	left of dur1.  if it is -1, then dur1 and dur2 has intersections
 *	and then dur1 is on the left of dur2.  if it is zero, then dur1
 *	and dur2 has intersections and then a duration contains the
 *	other.  if it is 1, then dur1 and dur2 has intersections and
 *	then dur1 is on the right on dur2.  if it is 2, then dur1 and
 *	dur2 has no intersections and dur1 is on the right of dur2.
 *	 closure is pointer to the Duration_t to output it.
 *	 intersection is pointer to pointer to the Duration_t to output
 *	it.
 *
 * returns:
 *	non zero	: illegal arguments
 *	zero		: normal end
 */
typedef enum duration_Inclusive_e {
	  Duration_has_no_intersect_dur1_is_on_left_of_dur2 = -2
	, Duration_has_intersect_dur1_is_on_left_of_dur2
	, Duration_has_intersect_dur_contains_the_other
	, Duration_has_intersect_dur1_is_on_right_of_dur2
	, Duration_has_no_intersect_dur1_is_on_right_of_dur2
} Duration_Inclusive_e;

extern int compare_durations(Duration_t			*dur1
				, Duration_t		*dur2
				, Duration_Inclusive_e	*cmp_result
				, Duration_t		*closure
				, Duration_t		*intersection
				);

/* This macro predicates the durations has intersections. */
#define Durations_has_intersections(Cmp_result) (		      \
		(Duration_has_no_intersect_dur1_is_on_left_of_dur2    \
						!= (Cmp_result))      \
		&& Duration_has_no_intersect_dur1_is_on_right_of_dur2 \
						!= (Cmp_result)	      \
	)

/* This macro predicates the dur1 is past of dur2. */
#define Duration1_is_past_of_duration2(Cmp_result)		\
						((int)(Cmp_result) < 0)

/* This macro predicates the dur2 is past of dur1. */
#define Duration2_is_past_of_duration1(Cmp_result)		\
						((int)(Cmp_result) > 0)

/* This macro predicates the durations are equal. */
#define Durations_are_equal(D1,D2)				\
				(memcmp((D1), (D2), sizeof(Duration_t)))

#ifdef __cplusplus
}
#endif
#endif /* } !defined(_DAYS_H_) */
